"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   03/10/2019

Command group for dataset exploration in ravenml.
"""

import click
import pydoc
from pkg_resources import iter_entry_points
from click_plugins import with_plugins
from colorama import Fore
from ravenml.utils.imageset import get_imageset_names, get_imageset_metadata
from ravenml.utils.dataset import get_dataset_names, get_dataset_metadata
from ravenml.utils.question import cli_spinner

# metedata fields to exclude when printing metadata to the user 
# these are specific to datasets at the moment
EXCLUDED_METADATA = ['filters', 'transforms', 'image_ids']

### OPTIONS ###
explore_details_opt = click.option(
    '-e', '--explore_details', 'explore_details', is_flag=True,
    help='Explore detailed metadata about all imagesets/datasets currently on S3 via a pager view.'
)

print_details_opt = click.option(
    '-p', '--print-details', 'print_details', is_flag=True,
    help='Print detailed metadata about all imagesets/datasets currently on S3 to the console.'
)

filter_details_opt = click.option(
    '-f', '--filter-details', 'filter_str', type=str,
    help=('Keyword/phrase filter for displaying detailed imageset/dataset metadata. '
            'Case sensitive and case insensitive filtering is performed.')
)


### COMMANDS ###
@click.group(help='Data exploration and dataset creation commands.')
@click.pass_context
def data(ctx):
    """Data exploration and dataset creation command group.
    
    Args:
        ctx (Context): click context object
    """
    pass

# TODO: eventually this will handle uploading/deleting the created
# dataset given by a plugin when create is called, see train.commands.process_result for example
@data.resultcallback()
@click.pass_context
def process_result(ctx: click.Context, result):
    pass


## Dataset Creation Commands ##
# TODO: determine interfaces for this command
@with_plugins(iter_entry_points('ravenml.plugins.data'))
@data.group(help='Create a new dataset.')
def create():
    click.echo("Eventually this will create datasets for you.")


## Imageset Commands ##
@data.command(help="List available image sets.")
@filter_details_opt
@explore_details_opt
@print_details_opt
def list_imagesets(print_details: bool, explore_details: bool, filter_str: str):
    """List available image sets on S3.
    
    Args:
        print_details (bool): T/F print detailed view to console
        explore_details (bool): T/F bring up detailed view in pager
        filter_str (str): string to detailed view on. None if not provided by user.
    """
    imageset_names = cli_spinner("Finding image sets on S3...", get_imageset_names)
    
    if explore_details or print_details:
        detailed_info = cli_spinner("Downloading imageset metadata from S3...", _get_detailed_imageset_info, imageset_names, filter_str=filter_str)
        if explore_details:
            pydoc.pager(detailed_info)      
        else:
            click.echo(detailed_info)
        return
        
    for name in imageset_names:
        click.echo(name)

@data.command(help='See detailed metadata about an image set.')
@click.argument('imageset_name')
def inspect_imageset(imageset_name: str):
    """See detailed metadata about a dataset.

    Args:
        dataset_name (str): string name of the dataset to inspect
    """
    try:
        metadata = cli_spinner("Downloading imageset metadata from S3...", get_imageset_metadata, imageset_name)
        # can check if the metadata returned is for an individual image if it has pose info
        if 'pose' in metadata.keys():
            click.echo(Fore.RED + ('Set-wide metadata not found. '
                                    'Falling back to sample image metadata.'))
        click.echo(_stringify_metadata(metadata, colored=True))
    # get_imageset_metadata will raise a value error if imageset name not found
    except ValueError as e:
        raise click.exceptions.BadParameter(imageset_name, param=imageset_name, param_hint='imageset name')
    # get imageset_metadata will raise a KeyError if the imageset does not contain any metadata files.
    except KeyError as e:
        raise click.exceptions.ClickException(f'Given imageset "{imageset_name}" does not contain any metadata files.')


## Dataset commands ##
@data.command(help='List available datasets.')
@filter_details_opt
@explore_details_opt
@print_details_opt
def list_datasets(print_details: bool, explore_details: bool, filter_str: str):
    """List available datasets.
    
    Args:
        print_details (bool): T/F print detailed view to console
        explore_details (bool): T/F bring up detailed view in pager
        filter_str (str): string to detailed view on. None if not provided by user.
    """
    dataset_names = cli_spinner("Finding datasets on S3...", get_dataset_names)

    if explore_details or print_details:
        detailed_info = cli_spinner("Downloading dataset metadata from S3...", _get_detailed_dataset_info, dataset_names, filter_str=filter_str)
        if explore_details:
            pydoc.pager(detailed_info)      
        else:
            click.echo(detailed_info)
        return
        
    for name in dataset_names:
        click.echo(name)

@data.command(help='See detailed metadata about a dataset.')
@click.argument('dataset_name')
def inspect_dataset(dataset_name: str):
    """See detailed metadata about a dataset.

    Args:
        dataset_name (str): string name of the dataset to inspect
    """
    try:
        metadata = cli_spinner("Downloading dataset metadata from S3...", get_dataset_metadata, dataset_name)
        click.echo(_stringify_metadata(metadata, colored=True))
    # get_dataset_metadata will raise a value error if the given dataset name does not exist on S3
    # OR if that dataset does not contain a metadata file. Currently these two situations are indistinguishible
    # TODO: separate errors for bad dataset name and dataset lacking metadata. This will need to be done in the 
    # get_dataset_metadata function
    except ValueError as e:
        raise click.exceptions.BadParameter(dataset_name, param=dataset_name, param_hint='dataset name')
        

### HELPERS ###
def _stringify_metadata(metadata: dict, colored=False) -> str:
    """Turn metadata into a nicely formatted string for displaying.

    Args:
        metadata (dict): dictionary of metadata
        colored (bool, optional): whether to colorize string, default False

    Returns:
        str: formatted metadata string
    """
    result = ''
    for key, val in metadata.items():
        if key not in EXCLUDED_METADATA:
            if colored:
                result += Fore.GREEN + str(key).upper() + ' ' + Fore.WHITE + str(val) + '\n'
            else:
                result += str(key).upper() + ' ' + str(val) + '\n'
    return result

def _get_detailed_dataset_info(datasets: list, filter_str:str=None) -> str:
    """Stringifies and concatenates metadata for a list of datasets.

    Args:
        datasets (list): list of dataset names
        filter_str (str, optional): string to filter metadata on

    Returns:
        str: concatenated and delimited metadata string for each dataset.
    """
    result = ''
    for dataset in datasets:
        try:
            metadata = get_dataset_metadata(dataset)
            str_metadata = _stringify_metadata(metadata)
            if filter_str:
                if filter_str in str_metadata:
                    result += str_metadata
                    result += '----------' '\n'
            else:
                result += str_metadata
                result += '----------' '\n'
        # we know we are only calling get_dataset_metadata on datsets that actually exist in S3,
        # so any ValueError indicates that dataset is missing metadata
        except ValueError as e:
            click.echo(f'Unable to find metadata in dataset "{dataset}", it will be skipped')
    return result

def _get_detailed_imageset_info(imagesets: list, filter_str:str=None) -> str:
    """Stringifies and concatenates metadata for a list of imagesets.

    Args:
        imagesets (list): list of imageset names
        filter_str (str, optional): string to filter metadata on

    Returns:
        str: concatenated and delimited metadata string for each imageset.
    """
    result = ''
    for imageset in imagesets:
        try:
            metadata = get_imageset_metadata(imageset)
            str_metadata = _stringify_metadata(metadata)
            if filter_str:
                # case sensitive and case insensitive checks
                if filter_str in str_metadata or filter_str.upper() in str_metadata or filter_str.lower() in str_metadata:
                    result += f'--IMAGESET NAME: {imageset.upper()}' + '\n'
                    result += str_metadata
                    result += '----------' + '\n'
            else:
                result += f'--IMAGESET NAME: {imageset.upper()}' + '\n'
                result += str_metadata
                result += '----------' + '\n'
        # we know we are only calling get_imageset_metadata on imagesets that actually exist in S3,
        # so we only need to check for KeyErrors (for imagesets that do not contain metadata files)
        except KeyError as e:
            click.echo(f'Unable to find metadata in imageset "{imageset}", it will be skipped')
    return result



import os
import boto3
import json
from pathlib import Path
from ravenml.utils.config import get_config
from ravenml.utils.local_cache import RMLCache

### DOWNLOAD FUNCTIONS ###
def list_top_level_bucket_prefixes(bucket_name: str):
    """Lists all top level prefixes in an S3 bucket.
    
    A top level prefix means it is the first in the chain. This will not list
    any subprefixes. 
    Ex: Bucket contains an element a/b/c/d.json, this function will only list a.
    
    Args:
        bucket_name (str): name of S3 bucket
        
    Returns:
        list: prefix strings
    """
    S3 = boto3.resource('s3')
    config = get_config()
    bucket_contents = S3.meta.client.list_objects(Bucket=bucket_name, Delimiter='/')
    contents = []
    if bucket_contents.get('CommonPrefixes') is not None:
        for obj in bucket_contents.get('CommonPrefixes'):
            contents.append(obj.get('Prefix')[:-1])
    return contents
    
def download_prefix(bucket_name: str, prefix: str, cache: RMLCache):
    """Downloads all files with the specified prefix into the provided local cache.

    Args:
        bucket_name (str): name of bucket
        prefix (str): prefix to filter on
        cache (RMLCache): cache to download files to
    
    Returns:
        bool: T if successful, F if no objects found
    """
    S3 = boto3.resource('s3')
    bucket = S3.Bucket(bucket_name)
    # filter bucket contents by prefix (add / to avoid accidental matching of invalid substrings
    # of valid dataset names)
    i = 0
    for obj in bucket.objects.filter(Prefix = prefix + '/'):
        # check to be sure object is not a folder by peeking at its last character
        i+=1
        if obj.key[-1] != '/':
            if not cache.subpath_exists(obj.key):
                subpath = os.path.dirname(obj.key)
                cache.ensure_subpath_exists(subpath)
                storage_path = cache.path / Path(obj.key)
                bucket.download_file(obj.key, str(storage_path))
    # T if objects were found and downloaded, F if not
    return i != 0

### UPLOAD FUNCTIONS ###
def upload_file_to_s3(prefix: str, file_path: Path, alternate_name=None):
    """Uploads file at given file path to model bucket on S3.

    Args:
        prefix (str): prefix for filename on S3
        file_path (Path): path to file
        alternate_name (str, optional): name to override local file name
    """
    S3 = boto3.resource('s3')
    config = get_config()
    model_bucket = S3.Bucket(config['model_bucket_name'])
    upload_path = prefix + '/' + file_path.name if alternate_name is None \
                    else prefix + '/' + alternate_name
    model_bucket.upload_file(str(file_path), upload_path)
        
def upload_dict_to_s3_as_json(s3_path: str, obj: dict):
    """Uploads given dictionary to model bucket on S3.

    Args:
        s3_path (str): full s3 path to save dictionary to, (no .json)
        obj (dict): dictionary to save
    """
    S3 = boto3.resource('s3')
    config = get_config()
    model_bucket = S3.Bucket(config['model_bucket_name'])   
    model_bucket.put_object(Body=json.dumps(obj, indent=2), Key=s3_path+'.json')
    


"""
Author(s):      Gavin Martin, Carson Schubert (carson.schubert14@gmail.com)
Data Created:   03/18/2019

Wrapper functions for prompting user input via questionary. Stolen from jigsaw.
"""

import sys
from halo import Halo
from questionary import prompt
from typing import Union

def in_test_mode() -> bool:
    """ Determines if we are running in an automated test or not. 
    This attribute is set via conftest.py in the ravenml/tests directory
    
    Returns:
        bool: true if in test mode, false otherwise
    """
    return hasattr(sys, "_called_from_test")

class Spinner:
    """Wrapper class to prevent bugs with Halo in pytest (see Issue #97)
    """

    def __init__(self, text, text_color):
        self.text = text
        if in_test_mode():
            return
        self._spinner = Halo(text=text, text_color=text_color)

    def start(self):
        if in_test_mode():
            return
        self._spinner.start()

    def succeed(self, text):
        if in_test_mode():
            return
        self._spinner.succeed(text=text)

def user_input(message: str, default="", validator=None) -> str:
    """Prompts the user for input
    
    Args:
        message (str): the message to give to the user before they provide
            input
        default (str, optional): Defaults to "". The default input response
        validator (Validator, optional): Defaults to None. A PyInquirer
            validator used to validate input before proceeding
    
    Returns:
        str: the user's response
    """
    if validator is not None:
        question = [
            {
                "type": "input",
                "name": 'value',
                "message": message,
                "default": default,
                "validate": validator,
            },
        ]
    else:
        question = [
            {
                "type": "input",
                "name": 'value',
                "message": message,
                "default": default
            },
        ]
    answer = prompt(question)
    return answer["value"]

def user_selects(message: str, choices: list, selection_type="list", sort_choices=True) -> Union[str, list]:
    """Prompts the user to select a choice(s) given a message
    
    Args:
        message (str): the message to give to the user before they choose
        choices (list): a list of the options to provide
        selection_type (str, optional): Defaults to "list". Should be "list"
            or "checkbox" for radio-button-style or checkbox-style selections
        sort_choices (bool, optional): Defaults to True. Whether to
            alphabetically sort the choices in the list provided to the user
    
    Returns:
        str or list: the str for the choice selected if "list" is the 
                        selection_type
                     the list of the choices selected if "checkbox" is the 
                        selection_type
    """
    question = [
        {
            "type": selection_type,
            "name": 'value',
            "message": message,
            "choices": sorted(choices) if sort_choices else choices
        },
    ]
    answer = prompt(question)
    return answer['value']
    
def user_confirms(message: str, default=False) -> bool:
    """Prompts the user to confirm an action by typing y/n
    
    Args:
        message (str): the message to give to the user before they choose
        default (bool, optional): Defaults to False.
    
    Returns:
        bool: the user's response in bool format
    """
    question = [
        {
            "type": "confirm",
            "name": "value",
            "message": message,
            "default": default
        },
    ]
    answer = prompt(question)
    return answer["value"]
    
def cli_spinner(text, func, *args, **kwargs):
    """ Halo spinner wrapper.

    Args:
        text (str): text to display while spinner is running
        func (function): function to execute while spinner runs
        *args (tuple, optional): ordered arguments required by func
        
    Raises:
        Automatically re-raises any exception raised by calling func after
        succeeding the spinner with a "Failed" tail
    """
    spinner = Spinner(text=text, text_color="magenta")
    spinner.start()
    try:
        result = func(*args,**kwargs)
    except Exception:
        spinner.succeed(text=text + 'Failed.')
        raise
    spinner.succeed(text=text + 'Complete.')
    return result
    



from setuptools import setup, find_packages
from os import path

proj_dir = path.abspath(path.dirname(__file__))
with open(path.join(proj_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ravenml',
    version='1.2',
    description='ML Training CLI Tool',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    license='MIT',
    packages=find_packages(),
    author='Carson Schubert, Abhi Dhir, Pratyush Singh',
    author_email='carson.schubert14@gmail.com',
    keywords= ['machine learning', 'data science'],
    download_url = 'https://github.com/autognc/ravenML/archive/v1.2.tar.gz',
    install_requires=[
        'Click>=7.0',
        'click-plugins>=1.0.4',
        'questionary>=1.0.2',
        'boto3>=1.9.86', 
        'shortuuid>=0.5.0',
        'halo>=0.0.26'
        'colorama>=0.3.9',
        'pyaml>=19.4.1',
    ],
    tests_require=[
        'pytest',
        'moto'
    ],
    entry_points='''
      [console_scripts]
      ravenml=ravenml.cli:cli
    ''',
)
      



"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   02/25/2019

Standard options which should be used by ravenml training plugins.
"""

import click
from ravenml.train.interfaces import TrainInput

### HELPERS/CALLBACKS ###

### OPTIONS ###

### PASS DECORATORS ###
"""Click pass decorator for use in training plugin commands that expect
to receive a TrainInput object. Triggers the creation of a TrainInput at ctx.obj
if one does not already exist there when the command executes. This behavior is used
to allow --help and other non-training plugin subcommands to run without triggering
TrainInput construction. See the TrainInput __init__ for more details.
"""
pass_train = click.make_pass_decorator(TrainInput, ensure=True)



"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   03/13/2019

Utility module for managing Jigsaw created datasets.
"""

import json
import boto3
from pathlib import Path
from botocore.exceptions import ClientError
from ravenml.utils.local_cache import RMLCache
from ravenml.utils.config import get_config
from ravenml.utils.aws import list_top_level_bucket_prefixes

imageset_cache = RMLCache('imagesets')
# name of config field
BUCKET_FIELD = 'image_bucket_name'

### PUBLIC METHODS ###
def get_imageset_names() -> list:
    """Retrieves the names of all available imagesets in bucket pointed to by global config.

    Returns:
        list: imageset names
    """
    config = get_config()
    return list_top_level_bucket_prefixes(config[BUCKET_FIELD])

def get_imageset_metadata(name: str, no_check=False) -> dict:
    """Retrieves imageset metadata. Downloads from S3 if necessary.

    Args:
        name (str): string name of imageset
        no_check (bool, optional): whether to verify metadata existence

    Returns:
        dict: imageset metadata
        
    Raises:
        ValueError: if given imageset name is invalid
    """
    if not no_check:
        try:
            _ensure_metadata(name)
        except ClientError as e:
            raise ValueError(name) from e
        except StopIteration as e:
            raise KeyError(name) from e
    return json.load(open(imageset_cache.path / Path(name) / 'metadata.json'))

# NOTE: this function is left here as a template for the eventual "get_imageset" function
# not implemented yet because we may find a better way to get image sets than actually downloading them locally
# def get_dataset(name: str) -> Dataset:
#     """Retrives a dataset. Downloads from S3 if necessary.

#     Args:
#         name (str): string name of dataset
    
#     Returns:
#         Dataset: dataset itself
        
#     Raises:
#         ValueError: if dataset name is invalid (re raised)
#     """
#     try:
#         _ensure_dataset(name)
#         return Dataset(name, get_dataset_metadata(name, no_check=True), dataset_cache.path / Path(name))
#     except ValueError:
#         raise
 

### PRIVATE HELPERS ###
def _ensure_metadata(name: str):
    """Ensure imageset metadata exists.
    NOTE: This function works around the fact that we don't have
    imageset wide metadata files for all imagesets. In that case, it
    picks a random metadata file and reports the fields contained within.

    Args:
        name (str): name of imageset
        
    Raises:
        ClientError: If the given imageset name does not exist in the S3 bucket.
        StopIteration: If the given imageset does not have any metadata files named according to the standard scheme.
    """
    S3 = boto3.resource('s3')
    config = get_config()
    image_bucket = S3.Bucket(config[BUCKET_FIELD])
    cache_metadata_path = Path(name) / 'metadata.json'      # relative path inside the cache where metadata will go
    if not imageset_cache.subpath_exists(cache_metadata_path):
        imageset_cache.ensure_subpath_exists(name)
        # attempt to download set-wide metadata.json
        imageset_bucket_metadata_key = f'{name}/metadata.json'
        metadata_download_absolute_path = imageset_cache.path / cache_metadata_path
        try:
            # attempt to grab imageset-wide metadata
            image_bucket.download_file(imageset_bucket_metadata_key, str(metadata_download_absolute_path))
        except ClientError as e:
            # fallback to grabbing a single image metadata file (better than nothing)
            prefix = f'{name}/meta_'
            # get all items in bucket with this prefix, but limit results to 1
            image_metadata_key_collection = image_bucket.objects.filter(Delimiter='/', Prefix=prefix).limit(1)
            # filter() returns a collection iterable, which we must convert to an iterator (generator) and call next on
            try:
                image_metadata_key = next(iter(image_metadata_key_collection)).key
                image_bucket.download_file(image_metadata_key, str(metadata_download_absolute_path))
            # explicitly reraise these errors for verbosity
            except ClientError as e:
                raise
            except StopIteration as e:
                raise

# NOTE: this function is left here as a template for the eventual "ensure_imageset" function
# not implemented yet because we may find a better way to get image sets than actually downloading them locally
# def _ensure_dataset(name: str):
#     """Ensures dataset exists.

#     Args:
#         name (str): name of dataset
        
#     Raises:
#         ValueError: if dataset name is invalid (no matching objects in S3 bucket)
#     """
    # config = get_config()
    # if not download_prefix(config[BUCKET_FIELD], name, dataset_cache):
    #     raise ValueError(name)
